/*
 * Copyright (c) 2009-2011 Yoshikazu Kuramochi
 * All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package ch.kuramo.javie.effects.stylize;

import java.nio.FloatBuffer;
import java.util.HashSet;
import java.util.Set;

import javax.media.opengl.GLUniformData;

import ch.kuramo.javie.api.IAnimatableBoolean;
import ch.kuramo.javie.api.IAnimatableInteger;
import ch.kuramo.javie.api.IShaderProgram;
import ch.kuramo.javie.api.IVideoBuffer;
import ch.kuramo.javie.api.VideoBounds;
import ch.kuramo.javie.api.IVideoBuffer.TextureFilter;
import ch.kuramo.javie.api.annotations.Effect;
import ch.kuramo.javie.api.annotations.Property;
import ch.kuramo.javie.api.annotations.ShaderSource;
import ch.kuramo.javie.api.annotations.Effect.Categories;
import ch.kuramo.javie.api.services.IConvolutionSupport;
import ch.kuramo.javie.api.services.IShaderRegistry;
import ch.kuramo.javie.api.services.IVideoEffectContext;
import ch.kuramo.javie.api.services.IVideoRenderSupport;

import com.google.inject.Inject;

@Effect(id="ch.kuramo.javie.Mosaic", category=Categories.STYLIZE)
public class Mosaic {

	@ShaderSource
	public static final String[] MOSAIC_SHARP_COLORS = {
		"uniform sampler2D texture;",
		"uniform vec2 blockSize;",			// テクスチャ座標系でのブロックサイズ
		"",
		"void main(void)",
		"{",
		"	gl_FragColor = texture2D(texture, (floor(gl_TexCoord[0].st/blockSize)+0.5)*blockSize);",
		"}"
	};


	@Property(value="10", min="1")
	private IAnimatableInteger horizontalBlocks;

	@Property(value="10", min="1")
	private IAnimatableInteger verticalBlocks;

	@Property
	private IAnimatableBoolean sharpColors;


	private final IVideoEffectContext context;

	private final IVideoRenderSupport support;

	private final IConvolutionSupport convolution;

	private final IShaderProgram sharpColorsProgram;

	@Inject
	public Mosaic(
			IVideoEffectContext context, IVideoRenderSupport support,
			IConvolutionSupport convolution, IShaderRegistry shaders) {

		this.context = context;
		this.support = support;
		this.convolution = convolution;
		sharpColorsProgram = shaders.getProgram(Mosaic.class, "MOSAIC_SHARP_COLORS");
	}

	public IVideoBuffer doVideoEffect() {
		IVideoBuffer input = context.doPreviousEffect();

		VideoBounds bounds = input.getBounds();
		if (bounds.isEmpty()) {
			return input;
		}

		int hBlocks = context.value(horizontalBlocks);
		int vBlocks = context.value(verticalBlocks);

		int w = bounds.width;
		int h = bounds.height;

		hBlocks = Math.min(hBlocks, w);
		vBlocks = Math.min(vBlocks, h);

		if (hBlocks == w && vBlocks == h) {
			return input;
		}

		boolean sharpColors = context.value(this.sharpColors);

		IVideoBuffer output = null;
		try {
			if (sharpColors) {
				output = doSharpColorsMosaic(input, hBlocks, vBlocks);
			} else {
				output = doAverageColorsMosaic(input, hBlocks, vBlocks);
			}
			return output;

		} finally {
			if (input != output) input.dispose();
		}
	}

	private IVideoBuffer doSharpColorsMosaic(IVideoBuffer input, int hBlocks, int vBlocks) {
		// テクスチャ座標系でのブロックサイズ
		float[] blockSize = new float[] { 1f/hBlocks, 1f/vBlocks };

		Set<GLUniformData> uniforms = new HashSet<GLUniformData>();
		uniforms.add(new GLUniformData("texture", 0));
		uniforms.add(new GLUniformData("blockSize", 2, FloatBuffer.wrap(blockSize)));

		return support.useShaderProgram(sharpColorsProgram, uniforms, null, input);
	}

	private IVideoBuffer doAverageColorsMosaic(IVideoBuffer input, int hBlocks, int vBlocks) {
		IVideoBuffer buffer = null;
		try {
			buffer = context.createVideoBuffer(new VideoBounds(hBlocks, vBlocks));
			average(input, buffer);
			scale(buffer, input);
			return input;
		} finally {
			if (buffer != null) buffer.dispose();
		}
	}

	private void average(IVideoBuffer src, IVideoBuffer dst) {
		IVideoBuffer srcOrig = src;
		TextureFilter srcFilter = src.getTextureFilter();
		IVideoBuffer tmpBuf = null;
		try {
			VideoBounds srcBounds = src.getBounds();
			VideoBounds dstBounds = dst.getBounds();
			int hBlocks = dstBounds.width;
			int vBlocks = dstBounds.height;

			double bwidth = (double)srcBounds.width / hBlocks;
			double bheight = (double)srcBounds.height / vBlocks;

			// ブロックサイズが5より大きい場合は複数回にわけて処理する
			// (大きなカーネルで一度に畳み込むと負荷が大きい。また、IConvolutionSupportの実装はカーネルサイズの上限が101(縦横掛けた値で))
			if (bwidth > 5 || bheight > 5) {
				tmpBuf = context.createVideoBuffer(new VideoBounds(
						(bwidth > 5) ? hBlocks*5 : hBlocks, (bheight > 5) ? vBlocks*5 : vBlocks));
				average(src, tmpBuf);
				src = tmpBuf;

				srcBounds = src.getBounds();
				bwidth = (double)srcBounds.width / hBlocks;
				bheight = (double)srcBounds.height / vBlocks;
			}

			int kwidth = (int)bwidth;
			int kheight = (int)bheight;
			int ksize = kwidth * kheight;
			float[] kernel = new float[ksize];
			float[] offset = new float[ksize*2];

			for (int j = 0; j < kheight; ++j) {
				for (int i = 0; i < kwidth; ++i) {
					int k = j*kwidth+i;
					kernel[k] = 1f/ksize;
					offset[k*2  ] = (float)((bwidth*((0.5+i)/kwidth-0.5)) / srcBounds.width);
					offset[k*2+1] = (float)((bheight*((0.5+j)/kheight-0.5)) / srcBounds.height);
				}
			}

			final VideoBounds bounds = dstBounds;
			final double[][] texCoords = new double[][] { {0, 0}, {1, 0}, {1, 1}, {0, 1} };

			Runnable operation = new Runnable() {
				public void run() {
					support.ortho2D(bounds);
					support.quad2D(bounds, texCoords);
				}
			};

			src.setTextureFilter(TextureFilter.LINEAR);
			convolution.convolve(src, dst, kernel, offset, operation, 0);

		} finally {
			if (tmpBuf != null) tmpBuf.dispose();
			if (src == srcOrig) src.setTextureFilter(srcFilter);
		}
	}

	private void scale(IVideoBuffer src, IVideoBuffer dst) {
		final VideoBounds dstBounds = dst.getBounds();

		final double[][] texCoords = new double[][] { {0, 0}, {1, 0}, {1, 1}, {0, 1} };

		Runnable operation = new Runnable() {
			public void run() {
				support.ortho2D(dstBounds);
				support.quad2D(dstBounds, texCoords);
			}
		};

		support.useFramebuffer(operation, 0, dst, src);
	}

}
